// Copyright 2016 Google Inc. All Rights Reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
//   Unless required by applicable law or agreed to in writing, software
//   distributed under the License is distributed on an "AS IS" BASIS,
//   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//   See the License for the specific language governing permissions and
//   limitations under the License.

#include <iostream>
#include <string>

#include "cctz/civil_time.h"
#include "cctz/time_zone.h"

int main() {
  cctz::time_zone lax;
  load_time_zone("America/Los_Angeles", &lax);

  // Time Programming Fundamentals @cppcon
  const auto tp = cctz::convert(cctz::civil_second(2015, 9, 22, 9, 0, 0), lax);

  cctz::time_zone nyc;
  load_time_zone("America/New_York", &nyc);

  std::cout << cctz::format("Talk starts at %H:%M:%S %z (%Z)\n", tp, lax);
  std::cout << cctz::format("Talk starts at %H:%M:%S %z (%Z)\n", tp, nyc);
}
